package CPU.instructions.LSL;

/**
 * The LSL instruction where data to be shifted is in a data reg and the shift count is in another data reg.
 * 1110ccc1ss101rrr = 0xe128   
 * 
 * mmm = mode 010
 * rrr = data register to be shifted
 * ccc = dataregister containing the shift count (modulo 64)
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * 
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2014.
 * @version 1.01 revamped.
 * Based on Tony Headford's  m68k.cpu.instruction (C) Tony Headford, see his icence below
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class LSLreg implements InstructionSet
{
    
    private static final int B_MASK = 0x80;
    private static final int W_MASK = 0x8000;
    
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                           // byte, word and long
        {
            switch (sz)
            {
                case 0:                                         // shift lsb
                    base = 0xe128;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[reg]&0xff;                    // byte to be shifted
                            if (nbrShifts!=0)
                            {
                                boolean lastBitSet; 
                                if (nbrShifts<=8)
                                    lastBitSet= (cpu.data_regs[reg]&(1<<(8-nbrShifts)))!=0;
                                else
                                    lastBitSet = false;
                                cpu.data_regs[reg]&=0xffffff00;                 // get rid of ls byte
                                cpu.reg_sr&=0xffe0;                             // all status flags set to 0
                                if (nbrShifts<8)
                                {
                                    val <<= nbrShifts;
                                    val&=0xff;                                  // do not combine with next line!
                                    cpu.data_regs[reg]|=val;                        
                                }
                                else
                                    val=0;
                                
                                if (lastBitSet)
                                    cpu.reg_sr|=17;
                            }
                            else
                            {
                                cpu.reg_sr&=0xfff0;                             // X unaffected, C  cleared by shift count 0
                            }
                            if (val==0)
                                cpu.reg_sr|=4;                 
                            else if ((val&B_MASK)!=0)
                                cpu.reg_sr|=8;
                            
                            /* remove at next review
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[opcode&0x7]&0xff;// byte to be shifted
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&0x80;                   // highest bit before shift
                                val<<=1;                            // shift one bit up
                            }
                            val&=0xff;
                            cpu.data_regs[opcode&0x7]&=0xffffff00; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;    // and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if ((val&0x80)!=0)                      // ms bit set
                                cpu.reg_sr+=8;                 // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (nbrShifts!=0)                       // X flag is unaffected by 0 count shift, C is cleared by it
                            {
                                if (lastOut!=0)
                                    cpu.reg_sr|=17;            // set X + C flags
                                else
                                    cpu.reg_sr&=~16;          // clear X flag (C flag already cleared)
                            }
                            */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };  
                    break;
            
                case 1:                                                         // word sized shift
                    base = 0xe168;
                    i = new CPU.Instruction()
                    {
                        
                        
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[reg]&0xffff;                  // word to be shifted
                            if (nbrShifts!=0)
                            {
                                boolean lastBitSet; 
                                if (nbrShifts<=16)
                                    lastBitSet= (cpu.data_regs[reg]&(1<<(16-nbrShifts)))!=0;
                                else
                                    lastBitSet = false;
                                cpu.data_regs[reg]&=0xffff0000;                 // get rid of ls byte
                                cpu.reg_sr&=0xffe0;                             // all status flags set to 0
                                if (nbrShifts<16)
                                {
                                    val <<= nbrShifts;
                                    val&=0xffff;
                                    cpu.data_regs[reg]|=val;                        
                                }
                                else
                                    val=0;
                                
                                if (lastBitSet)
                                    cpu.reg_sr|=17;
                            }
                            else
                            {
                                cpu.reg_sr&=0xfff0;                             // X unaffected, C  cleared by shift count 0
                            }
                            if (val==0)
                                cpu.reg_sr|=4;                
                            else if ((val&W_MASK)!=0)
                                cpu.reg_sr|=8;
                            
                            /* remove at next review
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// word to be shifted
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&0x8000;                 // highest bit before shift
                                val<<=1;                            // shift one bit up
                            }
                            val&=0xffff;
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if ((val&0x8000)!=0)                      // ms bit set
                                cpu.reg_sr+=8;                 // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (nbrShifts!=0)                       // X flag is unaffected by 0 count shift, C is cleared by it
                            {
                                if (lastOut!=0)
                                    cpu.reg_sr|=17;            // set X + C flags
                                else
                                    cpu.reg_sr&=~16;          // clear X flag (C flag already cleared)
                            } */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                    };  
                    break;
                default:                                                        // long sized shift
                    base = 0xe1a8;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[reg];                  
                            if (nbrShifts!=0)
                            {
                                boolean lastBitSet; 
                                if (nbrShifts<=32)
                                    lastBitSet= (cpu.data_regs[reg]&(1<<(32-nbrShifts)))!=0;
                                else
                                    lastBitSet = false;
                                cpu.reg_sr&=0xffe0;                             // all status flags set to 0
                                if (nbrShifts<32)
                                    val <<= nbrShifts;  
                                else
                                    val=0;
                                cpu.data_regs[reg]=val;
                                if (lastBitSet)
                                    cpu.reg_sr|=17;
                            }
                            else
                            {
                                cpu.reg_sr&=0xfff0;                             // X unaffected, C  cleared by shift count 0
                            }
                            if (val==0)
                                cpu.reg_sr|=4;                
                            else if (val<0)
                                cpu.reg_sr|=8;
                            
                           /* remove at next review
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[opcode&0x7];// long word to be shifted
                            int lastOut=0;            // current state of highest bit
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&0x80000000;
                                val<<=1;                            // shift one bit up
                            }
                            cpu.data_regs[opcode&0x7]=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if (val<0)                      // ms bit set
                                cpu.reg_sr+=8;                 // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (nbrShifts!=0)                       // X flag is unaffected by 0 count shift, C is cleared by it
                            {
                                if (lastOut!=0)
                                    cpu.reg_sr|=17;            // set X + C flags
                                else
                                    cpu.reg_sr&=~16;          // clear X flag (C flag already cleared)
                            }
                            */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
            DisassembledOperand src = new DisassembledOperand("d" + ((opcode >> 9) & 0x07));
            DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
            return new DisassembledInstruction(address, opcode, "lsl" + sz.ext(), src, dst);
    }
}
