package CPU.instructions.ROL;

/**
 * The LSL instruction where data to be rotateed is in a data reg and the rotate count is in another data reg.
 * 1110ccc1ss111rrr = 0xe138   
 * 
 * mmm = mode 010
 * rrr = data register to be rotateed
 * ccc = dataregister containing the rotate count (modulo 64)
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 *  
 * @author and copyright for my code (c) Wolfgang Lenerz 2013 - 2020.
 * @version 1.01 speedup
 * 
 * Based on code by Tony Headford, see his licence in accompanying file
*/
import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class ROLreg implements InstructionSet
{   
    private static final int bMask = 0x80;
    private static final int wMask = 0x8000;
    private static final int lMask = 0x80000000;
    
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                           // byte, word and long 
        {
            switch (sz)
            {
                case 0:                                         // rotate lsb
                    base = 0xe138;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                           
                            int reg = opcode&0x7;
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often we rotate
                            int val=cpu.data_regs[reg]&0xff;                    // byte to be rotated
                            if (nbrRots!=0)
                            {
                                int rots = nbrRots & 7;
                                val =((val<< rots)|(val>> 8-rots))&0xff;
                                cpu.data_regs[reg]&=0xffffff00;                 // get rid of ls byte
                                cpu.data_regs[reg]|=val;                        // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X
                            if ((bMask & val)!=0)                               // ms bit set
                                cpu.reg_sr|=8;                                  // so set N 
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((nbrRots!=0) && (val&1)!=0)                     // if last bit rotated out !=0 (and no zero rotation count): 
                                cpu.reg_sr+=1;                                  // set C flag
                  /*  old, remove at next review
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often we rotate
                            int val=cpu.data_regs[opcode&0x7]&0xff;             // byte to be rotateed
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrRots;cnt++)
                            {
                                lastOut=val&0x80;                               // highest bit before rotate
                                val<<=1;                                        // rotate one bit up
                                if (lastOut!=0)
                                    val|=1;
                            }
                            val&=0xff;
                            cpu.data_regs[opcode&0x7]&=0xffffff00;              // get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;                     // and fill it in with new value
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X 
                            if ((bMask & val)!=0)                               // ms bit set
                                cpu.reg_sr|=8;                                  // so set N 
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((val&1)!=0)                                     // if last bit rotated out !=0 : 
                                cpu.reg_sr+=1;                                  // set C flag
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };
                    break;
               
                case 1:                                         // word sized rotate
                    base = 0xe178;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often we rotate
                            int val=cpu.data_regs[reg]&0xffff;                  // word to be rotated, ensure upper word = 0
                            if (nbrRots!=0)
                            {
                                int rots = nbrRots & 15;
                                val =((val<< rots)|(val>> 16-rots))&0xffff;
                                cpu.data_regs[reg]&=0xffff0000;                 // get rid of lower word
                                cpu.data_regs[reg]|=val;                        // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X
                            if ((wMask & val)!=0)                               // ms bit set
                                cpu.reg_sr|=8;                                  // so set N 
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((nbrRots!=0) && (val&1)!=0)                     // if last bit rotated out !=0 (and no zero rotation count): 
                                cpu.reg_sr+=1;                                  // set C flag
                            
                            /*  old, remove at next review
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we rotate
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// word to be rotateed
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrRots;cnt++)
                            {
                                lastOut=val&0x8000;                 // highest bit before rotate
                                val<<=1;                            // rotate one bit up
                                if (lastOut!=0)
                                    val|=1;
                            }
                            val&=0xffff;
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if ((val&0x8000)!=0)                      // ms bit set
                                cpu.reg_sr|=8;                 // so set N flag
                            else if (val==0)
                                cpu.reg_sr|=4;                 // set Z flag
                            if (nbrRots!=0  & lastOut!=0)
                                cpu.reg_sr|=1;             // set C flag
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                    };  
                    break;
                    
                default:                                        // long word rotate
                    base = 0xe1b8;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often werotate
                            int val=cpu.data_regs[reg];                         // lword to berotated
                            if (nbrRots!=0)
                            {
                                int rots = nbrRots & 31;
                                val =(val<< rots)|(val>>>32-rots);
                                cpu.data_regs[reg]=val;                         // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X
                            if (val<0)                               
                                cpu.reg_sr|=8;                                  // negative
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((nbrRots!=0) && (val&1)!=0)                     // if last bit rotated out !=0 (and no zero rotation count): 
                                cpu.reg_sr+=1;                                  // set C flag
                            
                            /*  old, remove at next review
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often we rotate
                            int val=cpu.data_regs[opcode&0x7];                  // long word to be rotateed
                            int lastOut=0;            // current state of highest bit
                            for (int cnt=0;cnt<nbrRots;cnt++)
                            {
                                lastOut=val&0x80000000;
                                val<<=1;                            // rotate one bit up
                                if (lastOut!=0)
                                    val|=1;
                            }
                            cpu.data_regs[opcode&0x7]=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if (val<0)                // ms bit set
                                cpu.reg_sr|=8;                 // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr|=4;                 // set Z flag
                            if (nbrRots!=0  & lastOut!=0)
                                cpu.reg_sr|=1;             // set C flag
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
           //reg mode
            DisassembledOperand src = new DisassembledOperand("d" + ((opcode >> 9) & 0x07));
            DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
            return new DisassembledInstruction(address, opcode, "rol" + sz.ext(), src, dst);
    }
}
