package drivers; 

import smsqmulator.Types;

/**
 * A socket for UDP transactions.
 * 
 * @author and copyright (c) Wolfgang Lenerz 2020.
 * 
 * @version 
 * 1.00 (creation)
 * 
 * Some notes
 * A - This implementation keeps track of the last packet received, and implements some form of buffering:
 * 1 - 
 *  The normal SMSQE traps might expect some buffering, e.g. if they only want a line ending in a line feed. The rest of
 *  the packet should remain to be available for another line feed...
 * 2 -  
    For a tftp server that works ONLY via DatagramSockets, I need to be able to obtain the peer address and port of the last 
    peer that sent a packet to the server (the SMSQE IP interface makes no provisions for "packets"). For obvious reasons, the 
    server's datagram socket is not connected, since any (new) client must be able to send it a DatagramPacket. Hence, I keep 
    track of the last packet I received which contains the connection info on the peer that sent it. 
    It's up to the server to make sure that it does an immediate ip_getpeername after receiving a packet and before trying to receive 
    a new one.
    In this implementation, IP_GETPEERNAME returns the address and port of the connected peer if any, else that of the last received
    packet, if any.
    This is a pretty stupid and unsafe way to do this (if several clients try to contact the server at nearly the same time, 
    this might not go well).
   B - Java has no concept of an unbound DatagramSocket. 
    These sockets are (practically) always bound, either explicitly or implicitly. This means that a "bind" will always fail. 
    Hence the "unbound property".
    
 */    
public class UDPSocket
{
    /* Status of a connection */
    public static enum CONN_STATUS {NOT_CONNECTED, CONNECTING, CONNECTED, LISTENING};
    
    private java.net.DatagramSocket socket;   
    private CONN_STATUS status;                                 // current connection status
    private int errorNumber;                                    // error number
    private java.net.DatagramPacket receivedPacket;             // last packet received.
    private int nextByteOut=-1;                                 // pointer to next byte to be taken out from the packet
    private byte[] buff = new byte[65530];
    private boolean isUdd;
    private boolean channelIsClosed;
    private boolean unBound;                                    // is true if the socket is supposed to be unbound.
 
    /**
     * Create a socket.
     * 
     * @param cpu the cpu used.
     */
    public UDPSocket(CPU.MC68000Cpu cpu) 
    {                                                           //  normal QL               TCP/UDP             S-Basic
        int openType=cpu.data_regs[3];                          // 0 = old exclusive    create simple socket    OPEN
                                                                // 1 = old shared       create & connect        OPEN_IN     connect =  set output port & IP address
                                                                // 2 = new exclusive    create & bind           OPEN_NEW =  bind to input port & address
        if (openType>2 || openType<0)
        {
            cpu.data_regs[0]=Types.ERR_IPAR;
            return;
        }
        this.receivedPacket=null; 
        String mname=smsqmulator.SmsqeString.readSmsqeString(cpu.readMemoryLong(cpu.addr_regs[7]+8),cpu);
        String device = mname.substring(0,4).toUpperCase();
        this.isUdd=device.equals("UDD_");
        if (mname.length()==4)
            mname="";
        else
            mname=mname.substring(4);                       // remove device part

        String []nameparts=null;
        if (!mname.isEmpty())
        {
            // udp_ip:port    "both parameters optional          
            nameparts=mname.split(":");                // [0] = host , [1] = port
        }
        int newtype=-1;
        if (this.isUdd)                                     // if UDD, switch the open type around
        {
            switch (openType)
            {
                case 0:
                    newtype = 0;
                    break;
                case 1:
                    newtype = 2;
                    break;
                case 2:
                case 3:
                    newtype = 1;
                    break;
            }
                    
            if (newtype==-1)
            {
                cpu.data_regs[0]=Types.ERR_IPAR;
                return;
            }
            openType = newtype;
        }

        if ((openType!=0) && (nameparts==null || nameparts.length!=2 || nameparts[0].length()<1|| nameparts[1].length()<1))
        {
            cpu.data_regs[0]=Types.ERR_IPAR;
            return;
        }   
        this.status=UDPSocket.CONN_STATUS.NOT_CONNECTED;
        try
        {
            switch (openType)
            {
                case 0:
                    this.socket=new java.net.DatagramSocket(null);  // this automatically binds to a random port
                    this.unBound=true;
                    break;
                case 1:                                         // connects - sets outport & out address with - acts as client
                     java.net.InetAddress ina=java.net.InetAddress.getByName (nameparts[0]);
                     int port=Integer.parseInt(nameparts[1]);
                     this.socket=new java.net.DatagramSocket(); // this automatically binds to a random port
                     this.socket.connect(ina,port);             // socket is now bound and connected
                     this.status=UDPSocket.CONN_STATUS.CONNECTED;
                     break;
                case 2:                                         // binds = set in port and in IP address - acts as server
                case 3:                                         // same as case 2
                    port=Integer.parseInt(nameparts[1]); 
                    ina=java.net.InetAddress.getByName((nameparts[0]));
                    this.socket=new java.net.DatagramSocket(port,ina);// binds the socket to this 
                    this.status=UDPSocket.CONN_STATUS.CONNECTED;
                    break;
                default: 
                    {
                        cpu.data_regs[0]=Types.ERR_IPAR;
                        return;
                    } 
            }
            cpu.data_regs[0]=0;
        }   
        catch(java.net.BindException e)
        {
            cpu.data_regs[0]=Types.ERR_FDIU;
        }
        catch (Exception e)
        {
            cpu.data_regs[0]=Types.ERR_IPAR;
        }
    }

    /**
     * Close the socket.
     */
    public void close()
    {
        if (this.socket!=null)
        {
            if (this.isUdd)
            {
                byte[] byt=new byte[0];
                sendPacket(byt,null);
            }
            this.socket.close();
        }
    }

    /**
     * Gets the connection status.
     * 
     * @return the connection status, one of UDPSocket.CONN_STATUS.
     */
    public UDPSocket.CONN_STATUS getConnectionStatus()
    {
        return this.status;
    }
        
         
    /**
     * Gets the last error from this socket.
     * 
     * @return the last error encountered by the socket.
     */
    public int getLastError()
    {
        return this.errorNumber;
    }
    
    /**
     * Sets the last error for the socket.
     * 
     * @param err the last error encountered by the socket when send/receiving,connecting/listening.
     */
    public void setLastError(int err)
    {
        this.errorNumber=err;
    }
  
    /**
     * Binds a socket. This means the local port and address are set.
     * 
     * @param cpu the smsqmulator.cpu.MC68000Cpu object.
     * 
     * @return 0 or SMSQ/E error code.
     */
    public int bind(CPU.MC68000Cpu cpu)
    {
        byte[]id=new byte[4];
        for (int i=0;i<4;i++)
        {
           id[i]=(byte)(cpu.readMemoryByte(cpu.addr_regs[2]+4+i));
        }
        try
        {
            java.net.InetAddress ina=java.net.InetAddress.getByAddress(id);
            int family = cpu.readMemoryWord(cpu.addr_regs[2]);
            int port = cpu.readMemoryWord(cpu.addr_regs[2]+2);
 //           if (port==69)
   //             port = 55555;
            this.socket.bind(new java.net.InetSocketAddress(ina,port)); // socket is now bound
            this.status=UDPSocket.CONN_STATUS.CONNECTED;
        } 
        catch (java.net.UnknownHostException e)
        {
            this.errorNumber=4;                         // error in java.net.InetAddress.getByAddress creation
            return Types.ERR_IPAR;
        } 
        catch (Exception e)
        {
            this.errorNumber=4;                         // this **should** never happen
            return Types.ERR_IPAR;
        }   
        this.errorNumber=0;
        return 0;
    }
    
    /**
     * Connects a socket - sets peer address with port and address 
     * 
     * @param cpu the smsqmulator.cpu.MC68000Cpu (A2 points to a sockaddr structure)
     * 
     * @return SMSQ/E error code or 0
     */
    public int connect(CPU.MC68000Cpu cpu)
    {
        this.errorNumber=0;
        byte[]id=new byte[4];
        boolean isNull = true;
        for (int i=0;i<4;i++)
        {
           id[i]=(byte)(cpu.readMemoryByte(cpu.addr_regs[2]+4+i));
           if (id[i]!= 0)
               isNull=false;
        }
        if (isNull)
        {
            this.socket.disconnect();
            this.status=UDPSocket.CONN_STATUS.NOT_CONNECTED;
        }
        else
        {
            try
            {
                java.net.InetAddress ina=java.net.InetAddress.getByAddress(id);
                int family = cpu.readMemoryWord(cpu.addr_regs[2]);
                int port = cpu.readMemoryWord(cpu.addr_regs[2]+2);
                if (this.socket==null)
                    this.socket=new java.net.DatagramSocket();
                this.socket.connect(ina, port);
            }
            catch (java.net.UnknownHostException | NullPointerException e)
            {
                this.errorNumber=4;                                 // error in java.net.InetAddress.getByAddress or socket creation
            }
            catch (java.lang.IllegalArgumentException  e)
            {
                this.errorNumber=5;                                 // error when creating java.net.InetSocketAddress;
            }
            catch (SecurityException e)
            {
                this.errorNumber=7;                                 // security error when creating the socket
            }
            catch (java.net.SocketException e)
            {
                this.errorNumber=8;
            }
            if (this.errorNumber!=0)
            {
                return Types.ERR_IPAR;
            }
            this.status=UDPSocket.CONN_STATUS.CONNECTED;
        }
        this.errorNumber=0;
        return 0;
    }
    
  
                    
    /*  Actual IO operations.---------------------------------------------------------------------------------
     */
    
    /**
     * Send bytes out over the port.
     * 
     * 
     * @param cpu the cpu to get data to be sent from, it lies at (A1), ndr of bytes to be sent in D2.L.
     * @param size nbr of bytes to send
     * @param socketAddress
     * 
     * @return error code (0=ok, negative = SMSQ/E error code).
     */
    public int send(CPU.MC68000Cpu cpu,int size,java.net.InetSocketAddress socketAddress)
    {
        if (size==0)
            return 0;
        byte[] byt = new byte[size];                    // nbr of bytes to send
        int A1 = cpu.addr_regs[1];                      // where to get bytes from
        if ((A1&1)==0)
            fastRead(cpu,byt,A1,size);
        else
        {
            for (int i=0;i<size;i++)
            {
                byt[i]=(byte)(cpu.readMemoryByte(A1++));
            }
        }
        return sendPacket(byt,socketAddress);
    }
    /**
     * Send one packet
     * @param byt the bytes to send in the packet.
     * 
     * @return nbr of bytes sent (always length of byt= or SMSQE error code.
     */
    private int sendPacket (byte[]byt,java.net.InetSocketAddress socketAddress)
    {
        try
        {
            if (this.socket==null)
                return 0;
            java.net.DatagramPacket pcket;
            if (socketAddress!=null)
                pcket =new java.net.DatagramPacket(byt,byt.length,socketAddress);
            else
                pcket =new java.net.DatagramPacket(byt,byt.length);
            this.socket.send(pcket);
        }
        catch (java.io.IOException e)
        {
            this.errorNumber=10;
            return Types.ERR_TRNS;
        }
        catch (Exception e)
        {
            this.errorNumber=8;
            return Types.ERR_TRNS;
        }
        return byt.length;
    }
    
    /**
     * Send one byte out over the socket.
     * 
     * @param byt the byte to send
     * 
     * @return 0 or SMSQ/E error.
     */
    public int send1Byte(byte byt)
    {
        byte[] byta = new byte [1];
        byta[0]= byt;
        return sendPacket(byta,null);
    }

    /**
     * Implemnnt send to facility.
     * 
     * @param cpu
     * @param pointer
     * @param size
     * @return 
     */
    public int sendTo(CPU.MC68000Cpu cpu, int pointer, int size)    
    {
        byte[] id = new byte[4];
        for (int i=0;i<4;i++)
        {
           id[i]=(byte)(cpu.readMemoryByte(pointer+4+i));
        }
        try
        {
            return send (cpu,size,new java.net.InetSocketAddress(java.net.InetAddress.getByAddress(id),cpu.readMemoryWord(pointer+2)));
        }
        catch(java.net.UnknownHostException e)
        {
            this.errorNumber=4;
            return Types.ERR_TRNS;
        }
        catch (Exception e)
        {
            this.errorNumber=8;
            return Types.ERR_TRNS;
        }
    }
    
    /**
     * Read bytes from CPU memory and puts them into the byte array, only used if A1 is even!.
     * 
     * @param cpu  the cpu with the memory (as array of shorts).
     * @param byt  the array to be filled with the bytes read from memory.
     * @param A1   where to start reading memory content from.
     * @param size size of data to read
     */
    private void fastRead(CPU.MC68000Cpu cpu,byte []byt,int A1,int size)
    {
        int w;
        char[]mainMemory=cpu.getMemory();
        A1/=2;                                                  // index into memory array
        int i;
        for (i=0;i<size-1;i+=2)
        {
            w=mainMemory[A1++];                                 // get the word                          
            byt[i]=(byte)(w>>>8);                               // set upper byte
            byt[i+1]=(byte)(w&0xff);                            // set lower byte
        }
        if ((size&1)!=0)                                        // handle possible trailing byte
        {
            byt[i]=(byte)cpu.readMemoryByte(A1*2);
        }
    }
    
    private int getPacket()
    {
        this.errorNumber=0;
        int oldPtr = this.nextByteOut;                          // keep if peek;
       
        if (this.receivedPacket==null)
        {
            this.receivedPacket=new java.net.DatagramPacket(this.buff, this.buff.length);
        }
        if (this.nextByteOut==-1)
        {
            try
            {
                this.socket.setSoTimeout(10);             // wait for one second max
                this.socket.receive(this.receivedPacket);
    //            if (this.receivedPacket.getAddress()==null)
      //          {
        //            return Types.ERR_NC;
          //      }
                this.nextByteOut=0;
               // /*
                if (!this.socket.isConnected())
                {
                    this.socket.connect(this.receivedPacket.getAddress(), this.receivedPacket.getPort());
                }
                //*/
                if (this.isUdd  && this.receivedPacket.getLength()==0)
                    this.channelIsClosed = true;
            }
            catch (java.net.SocketTimeoutException e)           // the socket timed out before I got anything
            {
                this.errorNumber=11;                            // set specific error code
                return Types.ERR_NC;                            // show that my call was not complete.
            }
            catch (java.io.IOException e)                       // some other I/O error, eg. PortUnreachableException 
            {
                this.errorNumber=11;
                return Types.ERR_TRNS;                          // show it
            }
            catch (Exception e)                                 // catch all (e.g.IllegalBlockingModeException)
            {
                this.errorNumber=8;
                return Types.ERR_TRNS;
            }
        }
        return 0;
    }
    
    /**
     * Receive bytes from the port and copy them : either to (A1), updating A1 and D1, or into a destination byte array.on entry, D1 MAY hold a flag, which COULD be as follows
                  0x0             // no flag
	MSG_OOB		0x1		// process out-of-band data 
	MSG_PEEK	0x2		/* peek at incoming message 
	MSG_DONTROUTE	0x4		/* send without using routing tables 
	MSG_EOR		0x8		/* data completes record 
	MSG_TRUNC	0x10		/* data discarded before delivery 
	MSG_CTRUNC	0x20		/* control data lost before delivery 
	MSG_WAITALL	0x40		/* wait for full request or error 
 
 for the time being, only case 0 is honoured.
     *  
     * Does case 2 make sense?
     * 
     * @param cpu the CPU used.
     * @param size nbr of bytes to read
     * @param checkD1 if true, should check D1 for flags, else ignore D1.
     * @param destination if not null, a byte array into which to receive, if null, copy to cpumem at A1
     * 
     * @return SMSQ/E negative error code or the number of bytes I got
     * 
     * This implements some primitive buffereing, as follows:
     *   - if  there is no packet received, get one, using the timeout provided by SMSQE
     *      send as many bytes as required/possible to SMSQE, updating the next byte out pointer;
     *      if SMSQE aksed for less bytes than in the packet, do nothing more
     *      if SMSQE asked for more byte, send the one got set err.nc
     */
    public int receive(CPU.MC68000Cpu cpu,int size,boolean checkD1,byte[]destination)
    {   
        int errorCode=getPacket();
        if (errorCode!=0)
            return errorCode;
        int oldPtr = this.nextByteOut;
        int remaining = this.receivedPacket.getLength()-this.nextByteOut; // nbr of bytes available
        if (remaining < size)
            size = remaining;                                   // but I only want this many
        int limit = this.nextByteOut+size;                      // and this is how many I'll actually get
        int result;
         
        if (destination==null)                                  // write directly into (A1)
        {    
            int A1 = cpu.addr_regs[1];
            while(this.nextByteOut<limit)
            {
                cpu.writeMemoryByte(A1++, this.buff[this.nextByteOut++]);
            } 
            result = A1-cpu.addr_regs[1];
        }
        else
        {
            result = 0;
            while(this.nextByteOut<limit)
            {
                destination[result++]= this.buff[this.nextByteOut++];// fill in array
            }
        }
        if (checkD1)
            this.nextByteOut = oldPtr;                          // we were only peeking!
        if (this.nextByteOut >=this.receivedPacket.getLength()-1)
            this.nextByteOut=-1;                                // got all bytes from packet - there is no more
        return result;
    }
   
    /**
     * Receive bytes from the port and copy them to (A1),up to the first LF encountered.
     * 
     * @param cpu the CPU used.
     * @param size nbr of bytes to read
     * 
     * @return SMSQ/E negative error code or the number of bytes I got
     * (for buufer comments, see method just above)
     */
    public int receiveAndCheckForLF(CPU.MC68000Cpu cpu,int size)
    {
        int errorCode=getPacket();
        if (errorCode!=0)
            return errorCode;
        int remaining = this.receivedPacket.getLength()-this.nextByteOut; // nbr of bytes available
        if (remaining < size)
            size = remaining;                                   // but I only want this many
        int limit = this.nextByteOut+size;
        int result;
        byte currentByte;
        byte previousByte=0;
        byte LF = 10;                                           // LineFeed
        byte CR = 13;                                           // CarriageReturn
        int A1 = cpu.addr_regs[1];
        while(this.nextByteOut<limit)
        {
            currentByte = this.buff[this.nextByteOut++];  
            cpu.writeMemoryByte(A1++, currentByte);
            if (currentByte==LF)
            {
                if (previousByte == CR)
                    cpu.writeMemoryByte(--A1,LF);
                break;
            }
            previousByte = currentByte;
        } 
        result = A1-cpu.addr_regs[1];
        if (this.nextByteOut >=this.receivedPacket.getLength()-1)
            this.nextByteOut=-1;                                // got all bytes from packet - there is no more
        return result;
    }
    
    /**
     * Receive one byte from the port and return it.
     *  
     * for the time being, only cases 0 and 2 are honoured.
     * 
     * @param cpu the CPU used.
     * @param byt a byte array with ONE ELEMENT ONLY into which the byte goes, may not be null.
     * 
     * @return SMSQ/E error code or 0 if all ok.
     */
    public int receive1Byte(CPU.MC68000Cpu cpu,byte[]byt)
    {
        return receive(cpu,1,false,byt);
    }  
    
    /**
     * Implements receiveFrom (rcvfrm).
     * @param cpu
     * @param size
     * @return 
     */
    public int receiveFrom(CPU.MC68000Cpu cpu,int size)
    {   
        int errorCode=getPacket();
        if (errorCode!=0)
            return errorCode;
        
        int sockaddr = cpu.readMemoryLong(cpu.addr_regs[2]);
        cpu.writeMemoryWord(sockaddr, 2);
        cpu.writeMemoryWord(sockaddr+2, this.receivedPacket.getPort());
        java.net.InetAddress ineta = this.receivedPacket.getAddress();
        byte[] id =ineta.getAddress();
        int ipaddr=0;
        for (int i=0;i<4;i++)
        {
            ipaddr<<=8;
            ipaddr|=id[i]&0xff;
        }
        cpu.writeMemoryLong(sockaddr+4,ipaddr); 
        cpu.writeMemoryLong(sockaddr+8,0); 
        cpu.writeMemoryLong(sockaddr+12,0);
        int remaining = this.receivedPacket.getLength()-this.nextByteOut; // nbr of bytes available
        if (remaining < size)
            size = remaining;                                   // but I only want this many
        int limit = this.nextByteOut+size;                      // and this is how many I'll actually get
        int result;
        int A1 = cpu.addr_regs[1];
        while(this.nextByteOut<limit)
        {
            cpu.writeMemoryByte(A1++, this.buff[this.nextByteOut++]);
        } 
        result = A1-cpu.addr_regs[1];
        this.nextByteOut=-1;                                    // got all bytes from packet - there is no more
        return result;
    }
    
    /**
     * Get info on the remote host, put it at (A1).
     * at A1 lies : family, port, ipv4 address. On return D1.L = length of answer.
     * On entry, D2.L = length of buffer
     * 
     * @param cpu the cpu used
     * 
     * @return SMSQ/E error code (0 if no error).
     */
    public int getRemote(CPU.MC68000Cpu cpu)
    {
        java.net.InetAddress peerAddress=this.socket.getInetAddress();
        int peerPort;
        if (peerAddress==null)
        {
            if (this.receivedPacket==null)
            {
                cpu.data_regs[1]=0;
                this.errorNumber=10;
                return -7;
            }
            peerAddress=this.receivedPacket.getAddress(); 
            if (this.receivedPacket==null)                  // in thery that should never happen
            {
                cpu.data_regs[1]=0;
                this.errorNumber=10;
                return -7;
            }
            peerPort = this.receivedPacket.getPort();
        }
        else
             peerPort = this.socket.getPort();
        
        if (cpu.data_regs[2]<2)
            return Types.ERR_ORNG;
        cpu.writeMemoryWord(cpu.addr_regs[1], 2);
        
        if (cpu.data_regs[2]<4)
        {
            cpu.data_regs[1]=2;
            return 0;
        }
        cpu.writeMemoryWord(cpu.addr_regs[1]+2, peerPort);
        
        if (cpu.data_regs[2]<8)
        {
            cpu.data_regs[1]=4;
            return 0;
        }
        byte[] addr=peerAddress.getAddress();
        if (addr.length!=4)
        {
            cpu.data_regs[1]=0;
            this.errorNumber=10;
            return Types.ERR_NIMP;
        }
        int ttemp=0;
        for (int i=0;i<3;i++)
        {
            ttemp |=(addr[i]&0xff);
            ttemp<<=8;
        }
        ttemp |=(addr[3]&0xff);
        cpu.writeMemoryLong(cpu.addr_regs[1]+4, ttemp);
        cpu.data_regs[1]=8;
        return 0;
    }
  
    /**
     * Gets the name of the current socket, including address, port and connection status.
     * 
     * @return the full name, e.g "TCP_localport@local_address | peerport@peer_ip_addresst".
     */
    public String getName()
    {
        StringBuilder p=new StringBuilder (64);
        if (this.isUdd)
            p.append("UDD_");
        else
            p.append("UDP_");
        
        java.net.InetAddress ina = this.socket.getLocalAddress();
        if (ina!=null)
            p.append(ina.getHostAddress());
        p.append(":");
        int i = this.socket.getLocalPort();
        p.append(i!=-1?i:"0");
        p.append(" | ");
        
        ina = this.socket.getInetAddress();
        p.append(ina==null?"-":ina.getHostAddress());
        p.append(":");
        i = this.socket.getPort();
        p.append(i!=-1?i:"-");
        return p.toString();
    }
   
    private static String stripSite(String s)
    {
        if (s==null)
            return null;
        return (s.toLowerCase().endsWith(".site") && s.length()>5)?s.substring(0,s.length()-5):s;
    }

    /**
     * @return the isUDD
     */
    public boolean isUDD()
    {
        return this.isUdd;
    }
    
    /**
     * Checks whether the peer channel for a UDD channel is closed (i recieved an empty UDP packet).
     * 
     * @return true if peer channel is closed.
     */
    public boolean isClosed()
    {
        return this.channelIsClosed;
    }
    
    private boolean hasContent()
    {
        if (this.receivedPacket==null)
            return false;
        int remaining = this.receivedPacket.getLength() - this.nextByteOut;
        return remaining>0;
    }
    
     
    /**
     * Anything still pending?
     * @return 
     */
    public boolean isPending()
    {
        if (this.status!=UDPSocket.CONN_STATUS.CONNECTED)
            return false;                                       // nothing can be pending if I'm not connected
        if (hasContent())
            return true;
        getPacket();
        return hasContent();
    }
   
    
}
