package smsqmulator;
/**
 * Class to speed up some SMSQ/E floating point ari ops.
 * Some notes:
 * 
 * On the d1/d2 operations, d1 = mantissa, d2 - exponent.
 * 
 * tos=top of stack, pointed to by (a1)
 * nos=next on stack, pointed to by 6(a1)
 * 
 * SMSQ/E FPs have a 12 bit exponent, and a 31 bit mantissa + sign bit (bit 31 of the long word is the sign bit).
 * IEEE doubles (64 bits) "only" have an 11 bit exponent, rest is mantissa + sign bit (bit 63), so there might be some 
 * FPs that SMSQ/E can handle that IEEE can't : they are handed back to SMSQ/E.
 *  * 
 * THIS CLASS IS NOT THREAD SAFE!!!!!!!!!!!!!!!!!
 * Do not use one same object of this class by different threads concurrently.
 * 
 * @author and copyright (c) Wolfgang Lenerz 2015-2017
 * @version 
 * 1.10     sometimes we dont't need to round up the result (in double2QlFloat), sin(pi) returns 0
 * 1.09     sometimes we need to round up the result (in double2QlFloat)
 * 1.08     better handling of negative mantissa, -32768 is no longer a special case
 * 1.07     -32768 is treated as NAN and handed back to smsq/e.
 * 1.06     0x7b,0x86a80000 is a special value = 360. I'm sure this is a bug in SMSQ/e somehow (sbext_ext_turtle_asm)
 * 1.05     Cosmetic changes ("nega" variable definition removed).
 * 1.04     Better fp handling.
 * 1.03     (very slightly) faster ql to IEEE fp conversion, correct handling of some negative values.
 * 1.02     don't use c68 QL to IEEE routine, too many pbs (e.g.with $80e,$100000). 
 * 1.01     if nbrs don't fit in an IEEE double, or pose other pbs, use SMSQ/E routines.
 * 1.00     initial version.
 */
public class Arith
{
    private final static int exponentOne=0x3ff00000;            // IEEE exponent so that (2^exponent)=1
    private int oldPC;                                          // value of old program cunter
    private int whatOp;                                         // what operation we're using
    private final static double QL_2PI=6.2831853069365025;
    private final static double QL_PI =3.1415926534682512;
   
    /**
     * Handles the f.p. arithmetic operation.
     * 
     * @param cpu the cpu to handle the op for.
     * @param whatOp what operation to handle.
     */
    public final void handleOp(CPU.MC68000Cpu cpu,int whatOp)
    {
        this.oldPC=cpu.pc_reg;                                  // keep old Program Counter if error
        cpu.data_regs[0]=0;                                     // preset no error
        cpu.reg_sr|=4;                                          // and set mk68 status register accordingly : preset everything was ok
        cpu.pc_reg=cpu.readMemoryLong(cpu.addr_regs[7])/2;      // set new PC after operation
        cpu.addr_regs[7] += 4;                                  // already do rts
        this.whatOp=whatOp;                                     // keep this for possible error returns
        switch (whatOp)
        {
            case 0 :                                            // addd : d1/d2 + tos -> tos   
                double testVal=qlFloat2Double(cpu.data_regs[2]&0xffff, cpu.data_regs[1]);
                double2QlFloat(testVal+qlFloat2Double(cpu),cpu);
                break;
                
            case 1:                                             // add : tos + nos -> nos 
                testVal=qlFloat2Double(cpu);                    // tos
                cpu.addr_regs[1]+=6;                            // A1 is increased by 6 by this op
                double2QlFloat(testVal+qlFloat2Double(cpu),cpu);// get nos & do op
                break;
                
            case 2 :                                            // subd : tos - d1/d2 -> tos
                testVal=qlFloat2Double(cpu.data_regs[2]&0xffff, cpu.data_regs[1]);
                double2QlFloat(qlFloat2Double(cpu)-testVal,cpu);// tos - d1/d2
                break;
                
            case 3:                                             // sub :  nos - tos -> nos
                testVal=qlFloat2Double(cpu);                    // tos
                cpu.addr_regs[1]+=6;                            // point to nos
                double2QlFloat(qlFloat2Double(cpu)-testVal,cpu);// into what was nos
                break;
                
            case 4 :                                            // double : x -> 2x
                testVal=qlFloat2Double(cpu);
                double2QlFloat(testVal+testVal,cpu);
                break;
                
            case 5:                                             // halve : x -> x/2
                testVal=qlFloat2Double(cpu);
                double2QlFloat(testVal/2.0,cpu);
                break;
                
            case 6 :                                            // divd
                testVal=qlFloat2Double(cpu.data_regs[2]&0xffff, cpu.data_regs[1]);
                if (testVal==0)
                {
                    cpu.data_regs[0]=Types.ERR_OVFL;
                    cpu.reg_sr&=~4;
                }
                else
                {
                    double2QlFloat(qlFloat2Double(cpu)/testVal,cpu);
                }
                break;
                
            case 7:                                             // div : nos / tos ->nos
                testVal=qlFloat2Double(cpu);                    // tos
                if (testVal==0)
                {
                    cpu.data_regs[0]=Types.ERR_OVFL;
                    cpu.reg_sr&=~4;
                }        
                else
                {
                    cpu.addr_regs[1]+=6;
                    double2QlFloat(qlFloat2Double(cpu)/testVal,cpu);
                }
                break;
                
            case 8 :                                            // recip : x -> 1/x
                testVal=qlFloat2Double(cpu);
                if (testVal==0)
                {
                    cpu.data_regs[0]=Types.ERR_OVFL;
                    cpu.reg_sr&=~4;
                }
                else
                {
                    double2QlFloat(1.0/testVal,cpu);
                }
                break;
                
            case 9:                                             // muld : D1/D2 * TOS  -> TOS
                testVal=qlFloat2Double(cpu.data_regs[2]&0xffff, cpu.data_regs[1]);
                double2QlFloat(qlFloat2Double(cpu)*testVal,cpu);
                break;   
                
            case 10 :                                           // mul : tos*nos -> nos
                testVal=qlFloat2Double(cpu);
                cpu.addr_regs[1]+=6;
                double2QlFloat(qlFloat2Double(cpu)*testVal,cpu);
                break;
                
            case 11:                                            // square : x -> x*x
                testVal=qlFloat2Double(cpu);
                double2QlFloat(testVal*testVal,cpu);
                break;
                
            case 12 :                                           // square root
                double2QlFloat(Math.sqrt(qlFloat2Double(cpu)),cpu);   
                break;
                
            case 13:                                            // cos
                double2QlFloat(Math.cos(qlFloat2Double(cpu)),cpu);
                break;
                
            case 14 :                                           // sin
                double pp=qlFloat2Double(cpu);
                if (pp==QL_PI || pp==QL_2PI)
                    pp=0;
                else
                    pp=Math.sin(pp);
                double2QlFloat(pp,cpu);                                          
                break;
                
            case 15:                                            // cotan
                double2QlFloat(1.0/Math.tan(qlFloat2Double(cpu)),cpu);
                break;   
                
            case 16:                                            // tan
                double2QlFloat(Math.tan(qlFloat2Double(cpu)),cpu);
                break;
  /*            
            case 17: debug
                cpu.addr_regs[1]+=cpu.addr_regs[6];
                int what=cpu.readMemoryWord(cpu.addr_regs[0]);
                if (what==0x7fb)
                    double2QlFloat(Math.toRadians(qlFloat2Double(cpu)),cpu);// rad
                else
                    double2QlFloat(Math.toDegrees(qlFloat2Double(cpu)),cpu);// deg
                
                cpu.addr_regs[1]-=cpu.addr_regs[6];
                cpu.writeMemoryLong(cpu.addr_regs[6]+0x58, cpu.addr_regs[1]);
                cpu.data_regs[4]=2;
               // cpu.addr_regs[7]-=4;
            //    cpu.pc_reg=this.oldPC;   
                break;
      */          
            default: 
                cpu.data_regs[0]=Types.ERR_NIMP;
                cpu.reg_sr&=~4;                                 // what kind of op was that??????????????????????????!
                break;
        }
    }
    
    /**
     * Convert SMSQ/E float at (A1) into java double.
     * 
     * @param cpu
     * 
     * @return the double
     */
    private static double qlFloat2Double(CPU.MC68000Cpu cpu)
    {
        int exponent=cpu.readMemoryWord(cpu.addr_regs[1]);
        int mantissa=cpu.readMemoryLong(cpu.addr_regs[1]+2);   
        if (exponent==0x7b && mantissa==0x86a80000)
            return 360;
        return qlFloat2Double(exponent,mantissa);    
    }
     
    /**
     * Convert SMSQ/E float to java (IEEE) double.
     * Inspired partly by c68 routine. 
     * This gets the SMSQ/E mantissa and converts it into a double with an exponent of 1, so that the resulting value
     * is 1.mantissa. 1.0 is deducted from that and then this is multiplied by 2^(true exponent).
     * v1.03
     * 
     * @param exponent  SMSQ/E exponent
     * @param mantissa  SMSQ/E mantissa
     * 
     * @return the corresponding IEEE double,possibly Nan if out of range.
     */   
    private static double qlFloat2Double(int exponent, int mantissa)
    {
        if (mantissa==0)
            return 0;                                           // easy case 
        boolean neg=mantissa<0;                                                                                         
        
        if (neg)
        { 
            mantissa=-mantissa;
            exponent++;                                         // hmmm, why?, seems to work, though
        }
        else
            mantissa<<=1;                                       // remove SMSQ/E sign bit
        if (exponent <1027 || exponent>3070) 
            return Double.NaN;                                  // I can't handle that
        exponent-=0x800;                                        // SMSQ/E bias
        long lowerMantissa = (mantissa & 0xffff);               // keep
        lowerMantissa<<=20;                                     // lower part of the mantissa
        mantissa>>>=12;                                         // and upper bits
        mantissa|=Arith.exponentOne;                            // now combines special exponent + mantissa upper bits
        long interm1=mantissa;                                  // make into long &...
        interm1<<=32;                                           // shift into upper "integer" (long word)
        interm1|=lowerMantissa;                                 // and combine with lower mantissa parts
        double result= Double.longBitsToDouble(interm1) - 1.0;  // get the fractional value but deduct implied bit,i.e. 1
        return (neg)?-result*(Math.pow(2.0, (double)exponent)):result*(Math.pow(2.0, (double)exponent));
    }
    
    /**
     * Convert IEEE double back to SMSQ/E float and set it at (A1).
     * This converts the result of the operation back to an SMSQ/E float.
     * 
     * Based on C68 routine.
     * 
     * @param nbr the double to convert.
     * @param cpu
    */
    private void double2QlFloat(double nbr, CPU.MC68000Cpu cpu)
    {
       if (Double.isInfinite(nbr) || Double.isNaN(nbr))//||nbr == -32768.0)        // did the op give an error?
        {
            errorReturnAfterResult (cpu);                       // yes, do an error return to smsqe and have smsqe do the operation
            return;
        }
        long interm=Double.doubleToLongBits(nbr);
        int exponent, mantissa;
        if (nbr!=0)
        {
            mantissa=(int)interm;                               // lower long word mantissa
            interm>>>=32;                                       // higher long word
            exponent=(int)interm;
            boolean neg=exponent<0;
            exponent+=exponent;
            exponent>>>=1;
            if (exponent==0)
            {
                errorReturnAfterResult(cpu);
                return;
            }
            int temp=exponent;
            exponent>>>=20;
            exponent+=0x402;
            temp&=0xfffff;                                      // bits of the mantissa that were in higher long word lower 20 bits)
            temp|=0x100000;                                     // 
            temp<<=10;                                          // move into bits 30 and lower
            boolean temp1=(mantissa&0x380000)!=0;               // 1.08 do ne need to round up?
            mantissa>>>=22;
            mantissa|=temp;
            if (temp1 && mantissa!=0x7fffffff)                  // 1.08 yes, round up 1.10added 0x7ffffff
      //      if (temp1)                  // 1.08 yes, round up 1.10added 0x7ffffff
                 mantissa++;                                     
            if (neg)
            {
                if (mantissa != 0 && (mantissa & (mantissa-1)) == 0)// only one bit set in mantisa?
                {                   
                    mantissa<<=1;                                   // yes, normalize
                    mantissa|=0x80000000;                           
                    exponent-=1;
                }
                mantissa=0-mantissa;
            }
        }        
        else
        {
            exponent=0;
            mantissa=0;                                                     
        }
        cpu.writeMemoryWord(cpu.addr_regs[1], exponent);                   
        cpu.writeMemoryLong(cpu.addr_regs[1]+2, mantissa);                 
    }
    
    /**
     * For some reason, I couldn't really handle the operation (e.g. result of op = value outside an IEEE double).
     * Now return to SMSQ/E and do the operation there!
     * To do this, I must execute the instruction that originally sat at the (now) patched location.
     * 
     * @param cpu 
     */
    private void errorReturnAfterResult(CPU.MC68000Cpu cpu)
    {
        cpu.addr_regs[7]-=4;                                    // undo jump over rts
        switch (this.whatOp)
        {
            case 0 :                                            // addd 
            case 2 :                                            // subd : tos - d1/d2 -> tos
           // case 12 :                                           // square root
                resetCpu3Regs(cpu,0);
                break;
                
            case 1:                                             // add : 
                resetCpu3Regs(cpu,6);
                break;
                            
            case 3:                                             // sub :  nos - tos -> nos ****************
                cpu.pc_reg=this.oldPC;                          // actually do the jsr
                int s=this.oldPC *2+ cpu.readMemoryWordPCSignedInc();
                cpu.addr_regs[7] -= 4;
                cpu.writeMemoryLong(cpu.addr_regs[7],cpu.pc_reg*2);               
		cpu.pc_reg=(s&CPU.MC68000Cpu.CUT_OFF)/2;
               break; 
                
            case 4 :                                            // double : x -> 2x
                cpu.data_regs[0]=cpu.readMemoryWordSigned(cpu.addr_regs[1])&0xffff;// abridged move.w (a1),d0
                cpu.pc_reg =this.oldPC;                         // start at where we were 
                break;
                
            case 5:                                             // halve : x -> x/2
                int d=cpu.readMemoryWordSigned(cpu.addr_regs[1]);//  do a subq.w   #1,(a1)
                int r = d-1;
                cpu.writeMemoryWord(cpu.addr_regs[1],r);
                cpu.reg_sr&=0xffe0;                             // all flags 0
                boolean Sm = false;                             // source is never neg
                boolean Dm = (d & 0x8000) != 0;                 // dist is neg
                boolean Rm = (r & 0x8000) != 0;                 // result is neg
                //n z v c                                       // flags, x will be set to c
                //8 4 2 1
                if (r == 0)
                {
                    cpu.reg_sr+=4;                              // set Z flag
                }
                else if (Rm)
                {
                    cpu.reg_sr+=8;                              // set N flag
                }
                if (!Sm && Dm && !Rm)
                {
                    cpu.reg_sr+=2;                              // set V flag
                }
                if  (Rm && !Dm)
                {
                    cpu.reg_sr+=0x11;                           // set C & X flags
                }
                cpu.pc_reg =this.oldPC;                         // start at where we were 
                break;
                
            case 6 :                                            // divd
            case 9:                                             // muld : D1/D2 * TOS  -> TOS
            case 11:                                            // square : x -> x*x
                resetCpu5Regs(cpu,0);
                break;
                
            case 7:                                             // div : nos / tos ->nos
            case 10 :                                           // mul : tos*nos -> nos
                resetCpu5Regs(cpu,6);
                break;
                
            case 8 :                                            // recip : x -> 1/x
                cpu.addr_regs[1]-=2;                            // do a qa_push1
                cpu.writeMemoryWord(cpu.addr_regs[1], 1);
                cpu.addr_regs[1]-=4;
                cpu.writeMemoryLong(cpu.addr_regs[1], 0x08014000);
                cpu.pc_reg =this.oldPC+1;                       // start at where we were    
                break;
          
            case 12 :                                           // sqrt - do a move.l 2(a1),d0
                cpu.pc_reg=this.oldPC+1;                        // pc after that instruction
                cpu.reg_sr&=0xffe0;                             // all flags 0
                cpu.data_regs[0]=cpu.readMemoryLong(cpu.addr_regs[1]+2);
                if (cpu.data_regs[0]==0)
                    cpu.reg_sr+=4;                              // set Z flag
                else if(cpu.data_regs[0]<0)
                    cpu.reg_sr+=8;                              // set N flag
                break;
                
            case 13:                                            // cos
            case 14 :                                           // sin
                resetCpu4Regs(cpu);
                break;
                
            case 15:                                            // cotan
            case 16:                                            // tan
                s=cpu.readMemoryLong(cpu.addr_regs[1]+2);       // do a qa_dup
                cpu.addr_regs[1]-=4;
                cpu.writeMemoryLong(cpu.addr_regs[1], s);
                s=cpu.readMemoryWord(cpu.addr_regs[1]+4);                  
                cpu.addr_regs[1]-=2;
                cpu.writeMemoryWord(cpu.addr_regs[1], s);
                cpu.pc_reg =this.oldPC+1;                       // start at where we were  
                break;
        }
    }
          
    
    // One can imagine a more plaasing solution to the below than recopying (nearly) the same code three times, but none will be faster.
    /**
     * Switches back to SMSQ/E when d1/d2/d3/d4/d7 were saved on the stack.
     * 
     * @param cpu
     * @param sub if A1 was changed, sub this value from it.
     */
    private void resetCpu5Regs(CPU.MC68000Cpu cpu,int sub)
    {
        cpu.pc_reg =this.oldPC+1;                           // set new PC
        cpu.addr_regs[7] -= 4;                              // undo rts + prepare for D7
                                                            // d1/d2/d3/d4/d7
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[7]);
        cpu.addr_regs[7] -= 4;                              // prepare for D4
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[4]);
        cpu.addr_regs[7] -= 4;                              // prepare for D3
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[3]);
        cpu.addr_regs[7] -= 4;                              // prepare for D2
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[2]);
        cpu.addr_regs[7] -= 4;                              // prepare for D1
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[1]);

        cpu.addr_regs[1]-=sub;
        cpu.data_regs[0]=0;
    }
    
    /**
     * Switches back to SMSQ/E when d1/d2/d3 were saved on the stack.
     * 
     * @param cpu
     * @param sub if A1 was changed, sub this value from it.
     */
    private void resetCpu3Regs(CPU.MC68000Cpu cpu,int sub)
    {
        cpu.pc_reg =this.oldPC+1;                           // set new PC
        cpu.addr_regs[7] -= 4;                              // undo rts + prepare for D3
                                                            // d1/d2/d3
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[3]);
        cpu.addr_regs[7] -= 4;                              // prepare for D2
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[2]);
        cpu.addr_regs[7] -= 4;                              // prepare for D1
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[1]);

        cpu.addr_regs[1]-=sub;
        cpu.data_regs[0]=0;
    }
    
    /**
     * Switches back to SMSQ/E when d1/d2/d3/a2 were saved on the stack.
     * 
     * @param cpu
     * @param sub if A1 was changed, sub this value from it.
     */
    private void resetCpu4Regs(CPU.MC68000Cpu cpu)
    {
        cpu.pc_reg =this.oldPC+1;                           // set new PC
        cpu.addr_regs[7] -= 4;                              // undo rts + prepare for a2
    // d1/d2/d3/a2
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.addr_regs[2]);
        cpu.addr_regs[7] -= 4;                              // prepare for D3
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[3]);
        cpu.addr_regs[7] -= 4;                              // prepare for D2
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[2]);
        cpu.addr_regs[7] -= 4;                              // prepare for D1
        cpu.writeMemoryLong(cpu.addr_regs[7], cpu.data_regs[1]);
        cpu.data_regs[0]=0;
    }
}
