package smsqmulator;

/**
 * A fixed size FIFO queue that does not allow null elements. 
 * 
 * Null elements are never added to the buffer, but trying to do so will NOT generate any kind of error - they are just silently ignored.<p>
 * 
 * The buffer is backed by an array of element T. A buffer must have at least a capacity of 1 element.
 * 
 *  
 * 
 * @author and copyright (c) 2020 Wolfgang Lenerz.
 * @version 
 * 1.00
 * @param <T> the objects in the buffer.
 */
public class FifoQueue <T> 
{
    private final T[] buffer;                                   // the array backing this object
    private int tail;                                           // points to the last element inserted into the queue  
    private int head;                                           // point before the next element to be taken out of the queue
    private final int capacity;                                 // capacity of this queue
    
    /**
     * Creates the queue with a capacity of 10 elements.
     */
    public FifoQueue()
    {
        this (10);
    }

    /**
     * Creates the buffer with a capacity of n elements.
     * 
     * @param n the number of elements the buffer is to contain.
     * 
     * @throws IllegalArgumentException if the the number of elements the buffer is to have is smaller than 1.
     */
    @SuppressWarnings("unchecked")
    public FifoQueue (int n) throws IllegalArgumentException
    {
        if (n<1)
            throw new IllegalArgumentException();
        this.buffer = (T[]) new Object[n];
        this.capacity = n;
        this.tail = 0;
        this.head = 0;
    }

    /** 
     * Tries to add an element to the end of the queue.
     * 
     * @param newElement the element to add. <code>NULL</code> elements are NOT added to the buffer, they are silently discarded (no error, no exception).
     * 
     * @return true if element was added to the buffer, false if it wasn't, which indicates either that the queue is full, or that the element to be inserted was null.
     * 
     */
    public synchronized boolean add(T newElement) 
    {
        if (newElement==null)
            return false;
        this.tail++;                                            // where element is to be inserted
        if (this.tail==this.capacity)                           // if we reach the end switch back to start of buffer
            this.tail=0;
        if (this.tail == this.head)                             // we're overflowing
        {
           this.tail--;                                         // so refuse to add
           if (this.tail<0)
               this.tail=this.capacity-1;
           return false;
        }
        this.buffer[this.tail] = newElement;                    // insert element
        return true;
    }
    
    
    /**
     * Gets the head element of this buffer. This may be null if and only if the queue is empty.
     * 
     * @return the last element of this buffer. This will be null if and only if the queue is empty.
     */
    public synchronized T remove() 
    {
        if (this.head==this.tail)
            return null;                                        // queue is empty
        this.head++;
        if (this.head==this.capacity)
            this.head = 0;
        return this.buffer[this.head];
    }
    
    /**
     * Shows some info about the queue.
     * 
     * @return string with info (queue size).
     */
    @Override
    public String toString() 
    {
        return "FifoQueue (size= " + this.capacity+")";
    }   
}
