package smsqmulator;

/**
 * This handles TCP connections.
 * This is more complicated than Marcel's C code, since Java has three different objects for socket:
 * Socket (client side TCP socket), ServerSocket and DatagramSocket (UDP),and they are not interchangeable.
 * 
 * This class is not thread safe.
 * Only 50 connections can be open at once.
 * 
 * @author and copyright (c) wolfgang lenerz 2015-2019
 * 
 * @version 
 * 1.04 contains the error strings, class IPError is abandoned, sockets are in an array not a map ( a map with a small integer as key, what was I tinking?).
 *      writeCStringUnsafe used, better, nurt unsafe handling of  gethostbyname I/O op.
 * 1.03 IOB.FLIN, make D1, D2 comparison on (unsigned) word only, not LW.
 * 1.02 handleTrap added case 7.
 * 1.01 interface change for get_netname (case 6), needs at least 3.31.
 * 1.00 implements only SCK and TCP, trap3, 1-7,50,51,53,58,59,5b,5e,62,7b,7c. No UDP.
 * 0.00 initial version, just with trap#3;1-7.
 * 
 */
public class IPHandler
{
    private final IPSocket[]    sockets;        // array for keeping all sockets opened
    private static final int    MAX_SOCKETS=50;
    
    /**
     * These are the I/O codes I'm supposed to handle.
     */
    private static final int IOB_TEST   = 0x00;
    private static final int IOB_FBYT   = 0x01;
    private static final int IOB_FLIN   = 0x02;
    private static final int IOB_FMUL   = 0x03;
    private static final int IOB_SBYT   = 0x05;    
    private static final int IOB_SMUL   = 0x07;
    private static final int IOF_LOAD   = 0x48;
    private static final int IOF_SAVE   = 0x49;

    private static final int IP_LISTEN  = 0x50;
    private static final int IP_SEND    = 0x51;
 //   private static final int IP_SENDTO  = 0x52;
    private static final int IP_RECV    = 0x53;
    private static final int IP_RECVFM  = 0x54;
//    private static final int IP_GETOPT  = 0x55;
  //  private static final int IP_SETOPT  = 0x56;
 //   private static final int IP_SHUTDWN = 0x57;
    private static final int IP_BIND    = 0x58;
    private static final int IP_CONNECT = 0x59;
    private static final int IP_GETHOSTNAME = 0x5b;
  //  private static final int IP_GETSOCKNAME = 0x5c;
    private static final int IP_GETPEERNAME = 0x5d;
    private static final int IP_GETHOSTBYNAME= 0x5e;
  //  private static final int IP_GETHOSTBYADDR= 0x5f;
//    private static final int IP_SETHOSTENT       0x60
//    private static final int IP_ENDHOSTENT       0x61
    private static final int IP_H_ERRNO     = 0x62;
 /*
    private static final int IP_GETSERVBYNAME   = 0x64;
    private static final int IP_GETSERVBYPORT   = 0x65;
    private static final int IP_GETPROTOBYNAME  = 0x6e;
    private static final int IP_GETPROTOBYNUMBER= 0x6f;
    private static final int IP_INET_ATON   = 0x72;
    private static final int IP_INET_ADDR   = 0x73;
    private static final int IP_INET_NETWORK= 0x74;
    private static final int IP_INET_NTOA   = 0x75;
    private static final int IP_IOCTL       = 0x79;
    private static final int IP_GETDOMAIN   = 0x7a;
*/
    private static final int IP_H_STRERROR  = 0x7b;
    private static final int IP_ERRNO       = 0x7c;
   
    /**
     * Error strings, always in english.
     */ 
    public static final String[] IP_ERROR = 
    {
        "NO ERROR",
        "UNTYPED SOCKET - can't do I/O",
        "GENERIC SOCKET - can't do I/O",
        "SOCKET IS ALREADY CONNECTED - can't connect again",
        "WRONG HOST ADDRESS :- The host address is wrong or malformed",
        "WRONG PORT NUMBER",
        "I/O EXCEPTION DURING SOCKET CREATION",
        "SECURITY MANAGER VIOLATION DURING SOCKET CREATION",
        "UNKNOWN ERROR",
        "SOCKET IS NOT CONNECTED",
        "I/O EXCEPTION WHEN SENDING BYTES",
        "I/O EXCEPTION WHEN RECEIVING BYTES",
        "ILLEGAL SOCKET CONVERSION",
        "SERVER CHANNEL CLOSED",
        "FUNCTION NOT IMPLEMENTED",
        "Connection refused",
        "Socket already bound"
    };
    
    private int lastError;
    
    /**
     * Creates the object.
     */
    public IPHandler()
    {
        this.sockets=new IPSocket[MAX_SOCKETS];
    }
    
    /**
     * This is a very primitive way to find a free entry in the array.
     * @return a new free entry for a socket in the sockets array.
     */
    private int getNewKey()
    {
        for (int i=1;i<MAX_SOCKETS;i++)
        {
            if (this.sockets[i]==null)
                return i;
        }
        return -1;
    }
    
    /**
     * This handles all the SMSQE "trap" calls related to this device.
     * 
     * @param cpu  the cpu used.
     */
    public void handleTrap(CPU.MC68000Cpu cpu)
    {
        int d0;
        IPSocket socket=null;
                
        switch (cpu.data_regs[0])                               // what operation should we do?
        {
            case 1:                                             // open a socket or make an accept call
                socket=new IPSocket(cpu,this.sockets);          // try to open a socket (this changes cpu.data_regs[0]!)
                if (cpu.data_regs[0]==0)                        
                {    
                    int key = getNewKey();                      // new socket opened OK
                    if (key==-1)
                    {
                        cpu.data_regs[0]=Types.ERR_IMEM;
                    }
                    else
                    {
                        this.sockets[key]= socket;              // accept call
                        cpu.data_regs[7]=-key;                  // put negative key into D7
                    }
                }
                break;
                
            case 2:                                             // do I/O
                d0=cpu.readMemoryLong(cpu.addr_regs[7]);        // get IO key
                cpu.addr_regs[7]+=4;                            // adjust stack
                socket=this.sockets[-cpu.addr_regs[0]];         // get the socket that's being adderssed
                if (socket==null)
                    cpu.data_regs[0]=Types.ERR_ICHN;            // ooops
                else
                    handleIO (d0,cpu,socket);
                break;
                
            case 3:                                             // close socket
                socket=this.sockets[-cpu.addr_regs[0]];
                if (socket!=null)
                {
                    socket.close();                             // close it
                    this.sockets[-cpu.addr_regs[0]]=null;       // remove it from list
                }
                cpu.data_regs[0]=0;                             // show no error
                break;
                
            case 4:                                             // close all sockets, used at startup/reset
                for (int i =0;i<this.sockets.length;i++)
                {
                    if (this.sockets[i]!=null)
                    {
                        this.sockets[i].close();
                        this.sockets[i]=null;
                    }
                }
                cpu.data_regs[0]=0;
                break;
                
            case 5:                                             // get channel name into (a1), d2 = max length
                socket=this.sockets[-cpu.addr_regs[0]];
                String v;
                if (socket==null)
                    v="UNKNOWN SOCKET";
                else
                {
                    v=socket.getName();
                }
                
                int bytes=cpu.data_regs[2]&0xffff;
                smsqmulator.SmsqeString.writeSmsqeString(cpu.addr_regs[1],v,true,bytes,cpu);//write as many bytes as possible
                if (v.length()> bytes)
                    cpu.data_regs[0]=Types.ERR_IPAR;            // but signal buffer too small          
                else
                    cpu.data_regs[0]=0;                         // or signal all OK
                break;
                
            case 6:
                String s=getHName();                            // name of machine
                if (s==null)
                    s="Unknown name!";
                int l=s.length();
                int bufflen=cpu.readMemoryWord(cpu.addr_regs[1]+2);
                if (l>bufflen)
                {
                    cpu.data_regs[0]=Types.ERR_BFFL;
                }
                else
                {   
                    smsqmulator.SmsqeString.writeSmsqeString(cpu.readMemoryLong(cpu.addr_regs[1]+4), s, true, l,cpu);//write string
                    cpu.data_regs[0]=0;
                }
                break;
                
            case 7:
                try
                {
                    s=getLocalAddress();
                }
                catch (Exception u)
                { 
                    s="";
                }
                l=s.length();
                bufflen=cpu.readMemoryWord(cpu.addr_regs[1]+2);
                if (l>bufflen)
                {
                    cpu.data_regs[0]=Types.ERR_BFFL;
                }
                else
                {   
                    smsqmulator.SmsqeString.writeSmsqeString(cpu.readMemoryLong(cpu.addr_regs[1]+4), s, true, l,cpu);//write string
                    cpu.data_regs[0]=0;
                }    
                break;
            }
        
        if (cpu.data_regs[0]==0)
        {
            cpu.reg_sr|=4;
            this.lastError=0;
        }    
        else
        {
            cpu.reg_sr&=~4;                                     // set status reg according to D0;
            if (socket!=null)
                this.lastError=socket.getLastError();
        }
    }
    
    /**
     * This handles I/O.
     * 
     * The only real IO operations (where bytes are actually received/sent) allowed are : 
     *      test pending input, get 1 char, get 1 line ending in LF, get multiple bytes, scatter load
     *      send 1 byte, send multiple bytes, scatter save
     * D2 contains nbr of bytes or buffer remember for scatter load/save D2 is long, for other ops, D2 = word)
     * 
     * @param what type of operation
     * @param cpu the cpu used
     * @param m the socket this operation is to use
     */
    private void handleIO(int what,CPU.MC68000Cpu cpu,IPSocket socket)
    {
        int bytes;
        cpu.data_regs[0]=0;                                     // preset no error
        switch (what)
        {
           case IOB_TEST:                                       // test pending input
                if (socket.getSocketType()==IPSocket.TYPE_SCK)
                {
                    cpu.data_regs[0]=Types.ERR_EOF;             // simple unbound socket can't receive
                    break;
                }
                switch (socket.getConnectionStatus())
                {
                    case NOT_CONNECTED:
                    case CONNECTED:
                    case LISTENING:                             // for these three input could be pending
                        try
                        {
                            if (socket.rawIn.available()==0)
                                cpu.data_regs[0]=Types.ERR_NC;
                        }
                        catch (Exception e)
                        {
                            cpu.data_regs[0]=Types.ERR_NC;
                        }
                        break;
                        
                    default :
                        cpu.data_regs[0]=Types.ERR_NC;          // everything else nothing is pending.
                }
                break;
                
            case IOB_FBYT:                                      // get ONE char    
                byte[]byt=new byte[1];
                cpu.data_regs[0]= socket.receive1Byte(cpu,byt,cpu.data_regs[3]); 
                if (cpu.data_regs[0]==0)
                {
                    bytes=byt[0]&0xff;
                    cpu.data_regs[1]=(cpu.data_regs[1]&0xffffff00)|bytes;// return the byte in D1
                }
                break;
                
            case IOB_FLIN:                                      // get multiple bytes ending by LF into buffer at (a1), d2.w = buffer size
                int D2 = cpu.data_regs[2]&0xffff;               // size of buffer
                int D1W=cpu.data_regs[1]&0xffff;                // nbr of bytes already gotten in previous call
                if (D1W>=D2)                                    // if this happens, the buffer was filled w/o finding an LF
                {
                    cpu.data_regs[0]=Types.ERR_BFFL;            // signal no linefeed found
                    break;                                      // I'm done, I've got all I can get
                }
               
                byt=new byte[1];
                int A1 = cpu.addr_regs[1];                      // where to put bytes
                int i;
                for (i=D1W;i<D2;i++)
                {
                    cpu.data_regs[0]= socket.receive1Byte(cpu,byt,-1);  // try to get one byte
                    if (cpu.data_regs[0]!=0)
                        break;                                  // but didn't
                    cpu.writeMemoryByte(A1++, byt[0]);          // put byte
                    if (byt[0]=='\n')
                        break;                                  // done
                }
                if (cpu.data_regs[0]!=0 && cpu.data_regs[0]!=Types.ERR_NC)
                        break;                                  // genuine error, so leave
                if (i==D2 && byt[0]!='\n')
                    cpu.data_regs[0]=Types.ERR_BFFL;            // signal no line feed found
                    
                D2 = A1-cpu.addr_regs[1];                       // nbr of bytes read (this time around)
                cpu.addr_regs[1]=A1;
                cpu.data_regs[1]=cpu.data_regs[1]&0xffff0000;   // put nbr of bytes got so far into D1 lower word
                cpu.data_regs[1]|=(D2+D1W);
                break;
                
            case IOB_FMUL:                                      // get multiple bytes into buffer at (a1), d2.w = buffer size
                                                                // remember this may be called several times for one operation if NC
                D2=0xffff&cpu.data_regs[2];
                D1W=cpu.data_regs[1]&0xffff;
                if (D1W>=D2)
                    break;                                      // I'm done, I've got all I need
                bytes= socket.receive(cpu,D2,false,null);       // get bytes
                if (bytes<0)                                    // ooops
                {
                    cpu.data_regs[0]=bytes;                     // show error
                    break;                                      // and we're done
                }
                cpu.data_regs[1]+=bytes;
                cpu.addr_regs[1]+=bytes;                        // A1 pointer is increased by bytes just got
                cpu.data_regs[0]=D1W<D2?Types.ERR_NC:0;
                break;
                   
            case IOB_SBYT:                                      // send one byte, in D1.B
                cpu.data_regs[0]= socket.send1Byte(cpu,cpu.data_regs[1]);
                break;
                
            case IOB_SMUL:                                      // sned multiple bytes
                int D1=cpu.data_regs[1]&0xffff;                 // keep upper half of D1
                bytes = socket.send(cpu,cpu.data_regs[2]&0xffff);    // D2.w = amount to send         
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                    cpu.data_regs[1]=D1;                        // nothing was sent
                }
                else
                    cpu.data_regs[1]=D1|bytes;                  // that many bytes were sent
                break;
                
            case IOF_LOAD:                                      // get D2.L bytes from socket
                byt=new byte[cpu.data_regs[2]];
                while (cpu.data_regs[1]<cpu.data_regs[2])
                {
                    bytes= socket.receive(cpu,cpu.data_regs[2]-cpu.data_regs[1],false,byt); // try to get that many bytes
                    if (bytes<0)                                // ooops
                    {
                        cpu.data_regs[0]=bytes;                 // show error
                        break;                                  // and we're done
                    }
                    cpu.addr_regs[1]+=bytes;                    // A1 pointer is increased by bytes just got
                    cpu.data_regs[1]+=bytes;                    // nbr of bytes gotten
                }
                break;
                
            case IOF_SAVE:   
            case IP_SEND:                                       // send bytes to socket
                bytes = socket.send(cpu,cpu.data_regs[2]);      // D2.L = amount to send         
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                    cpu.data_regs[1]=0;                         // nothing was sent
                }
                else
                    cpu.data_regs[1]=bytes;                     // that many bytes were sent
                break;
                
            case IP_LISTEN:                                     // set a (server) socket to listen
                cpu.data_regs[0]= socket.listen(cpu);
                break;
             
     /*           
            case IP_SENDTO:
                cpu.data_regs[0]= m.sendTo(cpu,cpu.data_regs[2]);
                break;
       */         
            case IP_RECV:
                bytes= socket.receive(cpu,cpu.data_regs[2],true,null);
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;
                    cpu.data_regs[1]=0;
                }
                else
                    cpu.data_regs[1]=bytes;
                break;
                
            case IP_BIND :
                cpu.data_regs[0]=  socket.bind(cpu);
                break;        
        
            case IP_CONNECT:
                cpu.data_regs[0]=  socket.connect(cpu);
                break;
                
            case IP_GETHOSTNAME:
            {
                String localhostname = getHName();
                if (localhostname!=null)
                    writeCString (cpu,cpu.addr_regs[1],localhostname,cpu.data_regs[2]&0xfff);
                else
                    cpu.data_regs[0]=Types.ERR_FDNF;
            }
            break;  
                
            case IP_GETHOSTBYNAME:    
                /* ON EXIT
                 (Simplified) structure pointed to by A2

                0   pointer to A
                4   pointer to C
                8   long    type (2)
                12  long    4
                16   
                
                A       pointer to B
                B       pointer to inet address (4 bytes if ipv4)
                        B is followed by a long word of 0
                
                C       pointer to name
                */
                try
                {
                    int A2 = cpu.addr_regs[2];
                    String name =readCString(cpu,cpu.addr_regs[1]);
                    java.net.InetAddress addr=java.net.InetAddress.getByName(name);
                    
                    byte[] rawaddr=addr.getAddress();
                    cpu.writeMemoryLong(A2+8,2);                //type
                    cpu.writeMemoryLong(A2+12,4);     
                    int temp=A2+16;  
                    cpu.writeMemoryLong(A2,temp);               // pointer to A
                    cpu.writeMemoryLong(temp,temp+12);           // ptr to B into A
                    temp+=12;
                    cpu.writeMemoryLong(temp,temp+8);           // ptr to value into B
                    cpu.writeMemoryLong(temp+4,0);              // long word of 0 after B
                    bytes = temp+8;
                    temp=(rawaddr[0]&0xff)<<24 ;
                    temp+=((rawaddr[1]&0xff)<<16);
                    temp+=((rawaddr[2]&0xff)<<8);
                    temp+=(rawaddr[3]&0xff);
                    cpu.writeMemoryLong(bytes,temp);
                    temp=bytes + 4;                             // next free space
                    
                    bytes = writeCStringUnsafe (cpu,temp,name);
                    bytes+=1+temp;
                    bytes &=0xfffffffe;
                    
                    cpu.writeMemoryLong(A2+4,bytes);               // pointer to C
                    cpu.writeMemoryLong(bytes, temp);           // pointer to name
                } 
                catch (java.net.UnknownHostException ex)
                {
                    socket.setLastError (4);
                    cpu.data_regs[0]=Types.ERR_FDNF;
                }
                break;
                
            case IP_GETPEERNAME:
                cpu.data_regs[0]=socket.getRemote(cpu);
                break;
                
            case IP_H_ERRNO:
                cpu.data_regs[0]=0;
                cpu.data_regs[1]=socket.getLastError();
                break;
                
     //!!!!! this is a really unsafe operation: the address for the text to be returned is in A1 but NO MAXIMAL LENGTH IS GIVEN!!!! Stupid *!& C stuff       
            case IP_H_STRERROR:
                String s = IPHandler.IP_ERROR[socket.getLastError()];
                writeCStringUnsafe (cpu,cpu.addr_regs[1],s);
                cpu.data_regs[0]=0;
                break;
            
            case IP_ERRNO:
                cpu.data_regs[0]=0;
                cpu.data_regs[1]=this.lastError;
                break;
                
            default:
                socket.setLastError (14);
                cpu.data_regs[0]=Types.ERR_NIMP;                // what is that? whatever it is, it is not implemented for this device
                break;
        }
        this.lastError=socket.getLastError();
     }
    
    /**
     * Gets the socket.
     * 
     * @param key which socket to get.
     * 
     * @return the socket retrieved, null if not found.
     */
    public IPSocket getSocket(int key)
    {
        return this.sockets[key];
    }
    
     /**
     * Writes a null terminated string to memory without know how much mem I have.
     *
     * Note : This is an awful hack, but: Sometimes I DON'T KNOW THE LENGTH OF THE BUFFER I'M WRITING TO,
     * as the interface doesn't give me that information.
     * So I just use the length of the string (+1 for the terminating 0) as maximum size of the buffer!!!!
     * Hello possible buffer overflow. Just pray, I guess.
     * 
     * @param cpu the CPU used
     * @param address where to write to
     * @param s       the string to write 
     * @return nbr of bytes written
     */
    public static final int writeCStringUnsafe(CPU.MC68000Cpu cpu,int address,String s)
    {
        return writeCString(cpu, address,s,s.length()+1);
    }
    
    /**
     * Writes a null terminated string to (a buffer in) memory.
     *
     * @param cpu the CPU used
     * @param address where to write to
     * @param s       the string to write 
     * @param maxlength max length of buffer
     * @return nbr of bytes written
     */
    public static final int writeCString(CPU.MC68000Cpu cpu,int address,String s,int maxlength)
    {
        smsqmulator.SmsqeString.writeSmsqeString(address,s+"\0",false,maxlength,cpu);
        int bytes = s.length()+1;
        if (bytes>(maxlength))                                  // if string written was to big for my buffer, set last char to 0
        {                                                       // so that C has the 0 terminated string it expects
            bytes=maxlength;
            cpu.writeMemoryByte(address+bytes-1, 0);
            cpu.data_regs[0]=Types.ERR_OVFL;
        }
        else
        {
            cpu.data_regs[0]=0;
        }
        return bytes;
    }
    
    /**
     * Reads a "C" type string from memory and converts it to a Java String, with SMSQE char conversion.
     * 
     * Note: potentially, this might cause the StringBuilder to overflow if no "0" is found anywhere, but I'll get an out of mem bounds error first.
     * 
     * @param cpu   the cpu with the memory array.
     * @param address where to read string in the memory.
     * 
     * @return the corresponding String.
     */
    public static final String readCString(CPU.MC68000Cpu cpu,int address)
    {
        StringBuilder p=new StringBuilder(100);
        byte b;
        while((b=(byte)cpu.readMemoryByte(address++))!=0)
        {
            p.append(smsqmulator.SmsqeString.convertToJava(b));
        }
        return p.toString();
    }
    
    /**
     * Try to get the host name.
     * 
     * @return name or null if problem.
     */
    private static String getHName ()
    {
        try
        {
            return java.net.InetAddress.getLocalHost().getHostName();// try obvious solution
        }
        catch (Exception u)
        { /*NOP*/}                                              // if problem use code below

        String host = System.getenv("COMPUTERNAME");
        if (host != null)
            return stripSite(host);
        return stripSite(System.getenv("HOSTNAME"));
    }
    
    /**
     * Strips the word ".site" from the end of a string.
     * 
     * @param s 
     * 
     * @return intiial string stripped of the word ".site" at the end, if it was there.
     */
    private static String stripSite(String s)
    {
        if (s==null)
            return null;
        return (s.toLowerCase().endsWith(".site") && s.length()>5)?s.substring(0,s.length()-5):s;
    }
   
    
    /**
     * Get the local inet v4 address as string, without it being a loopback address.
     *  
     * @return 
     */
    private static String getLocalAddress() //throws java.net.SocketException
    {
        java.util.Enumeration<java.net.NetworkInterface> ifaces;
        try
        {
            ifaces= java.net.NetworkInterface.getNetworkInterfaces();
        }
        catch (Exception e)
        {
            return "";
        }
        while( ifaces.hasMoreElements() )
        {
            java.net.NetworkInterface iface = ifaces.nextElement();
            java.util.Enumeration<java.net.InetAddress> addresses = iface.getInetAddresses();
            while( addresses.hasMoreElements() )
            {
                java.net.InetAddress addr = addresses.nextElement();
                if( addr instanceof java.net.Inet4Address && !addr.isLoopbackAddress() )
                {
                    return addr.getHostAddress();
                }
            }
        }
        return "";
    }
}

        